from ._version import __version__

# required for Python 3.8 on Windows
import os
if hasattr(os, 'add_dll_directory'):
    if os.getenv('SCIPOPTDIR'):
        os.add_dll_directory(os.path.join(os.environ['SCIPOPTDIR'].strip('"'), 'bin'))

# export user-relevant objects:
from pyscipopt.Multidict import multidict as multidict
from pyscipopt.scip      import Model as Model
from pyscipopt.scip      import Variable as Variable
from pyscipopt.scip      import MatrixVariable as MatrixVariable
from pyscipopt.scip      import Constraint as Constraint
from pyscipopt.scip      import MatrixConstraint as MatrixConstraint
from pyscipopt.scip      import Benders as Benders
from pyscipopt.scip      import Benderscut as Benderscut
from pyscipopt.scip      import Branchrule as Branchrule
from pyscipopt.scip      import Nodesel as Nodesel
from pyscipopt.scip      import Conshdlr as Conshdlr
from pyscipopt.scip      import Eventhdlr as Eventhdlr
from pyscipopt.scip      import Heur as Heur
from pyscipopt.scip      import Presol as Presol
from pyscipopt.scip      import Pricer as Pricer
from pyscipopt.scip      import Prop as Prop
from pyscipopt.scip      import Reader as Reader
from pyscipopt.scip      import Sepa as Sepa
from pyscipopt.scip      import LP as LP
from pyscipopt.scip      import IISfinder as IISfinder 
from pyscipopt.scip      import PY_SCIP_LPPARAM as SCIP_LPPARAM
from pyscipopt.scip      import PY_SCIP_BASESTAT as SCIP_BASESTAT
from pyscipopt.scip      import readStatistics as readStatistics
from pyscipopt.scip      import Expr as Expr
from pyscipopt.scip      import MatrixExpr as MatrixExpr
from pyscipopt.scip      import MatrixExprCons as MatrixExprCons
from pyscipopt.scip      import ExprCons as ExprCons
from pyscipopt.scip      import quicksum as quicksum
from pyscipopt.scip      import quickprod as quickprod
from pyscipopt.scip      import exp as exp
from pyscipopt.scip      import log as log
from pyscipopt.scip      import sqrt as sqrt
from pyscipopt.scip      import sin as sin
from pyscipopt.scip      import cos as cos
from pyscipopt.scip      import PY_SCIP_RESULT          as SCIP_RESULT
from pyscipopt.scip      import PY_SCIP_PARAMSETTING    as SCIP_PARAMSETTING
from pyscipopt.scip      import PY_SCIP_PARAMEMPHASIS   as SCIP_PARAMEMPHASIS
from pyscipopt.scip      import PY_SCIP_STATUS          as SCIP_STATUS
from pyscipopt.scip      import PY_SCIP_STAGE           as SCIP_STAGE
from pyscipopt.scip      import PY_SCIP_NODETYPE        as SCIP_NODETYPE
from pyscipopt.scip      import PY_SCIP_PROPTIMING      as SCIP_PROPTIMING
from pyscipopt.scip      import PY_SCIP_PRESOLTIMING    as SCIP_PRESOLTIMING
from pyscipopt.scip      import PY_SCIP_HEURTIMING      as SCIP_HEURTIMING
from pyscipopt.scip      import PY_SCIP_EVENTTYPE       as SCIP_EVENTTYPE
from pyscipopt.scip      import PY_SCIP_LOCKTYPE        as SCIP_LOCKTYPE
from pyscipopt.scip      import PY_SCIP_LPSOLSTAT       as SCIP_LPSOLSTAT
from pyscipopt.scip      import PY_SCIP_BRANCHDIR       as SCIP_BRANCHDIR
from pyscipopt.scip      import PY_SCIP_BENDERSENFOTYPE as SCIP_BENDERSENFOTYPE
from pyscipopt.scip      import PY_SCIP_ROWORIGINTYPE as SCIP_ROWORIGINTYPE
from pyscipopt.scip      import PY_SCIP_SOLORIGIN as SCIP_SOLORIGIN
from pyscipopt.scip      import PY_SCIP_NODETYPE as SCIP_NODETYPE
from pyscipopt.scip      import PY_SCIP_IMPLINTTYPE as SCIP_IMPLINTTYPE



##@file expr.pxi
#@brief In this file we implemenet the handling of expressions
#@details @anchor ExprDetails <pre> We have two types of expressions: Expr and GenExpr.
# The Expr can only handle polynomial expressions.
# In addition, one can recover easily information from them.
# A polynomial is a dictionary between `terms` and coefficients.
# A `term` is a tuple of variables
# For examples, 2*x*x*y*z - 1.3 x*y*y + 1 is stored as a
# {Term(x,x,y,z) : 2, Term(x,y,y) : -1.3, Term() : 1}
# Addition of common terms and expansion of exponents occur automatically.
# Given the way `Expr`s are stored, it is easy to access the terms: e.g.
# expr = 2*x*x*y*z - 1.3 x*y*y + 1
# expr[Term(x,x,y,z)] returns 1.3
# expr[Term(x)] returns 0.0
#
# On the other hand, when dealing with expressions more general than polynomials,
# that is, absolute values, exp, log, sqrt or any general exponent, we use GenExpr.
# GenExpr stores expression trees in a rudimentary way.
# Basically, it stores the operator and the list of children.
# We have different types of general expressions that in addition
# to the operation and list of children stores
# SumExpr: coefficients and constant
# ProdExpr: constant
# Constant: constant
# VarExpr: variable
# PowExpr: exponent
# UnaryExpr: nothing
# We do not provide any way of accessing the internal information of the expression tree,
# nor we simplify common terms or do any other type of simplification.
# The `GenExpr` is pass as is to SCIP and SCIP will do what it see fits during presolving.
#
# TODO: All this is very complicated, so we might wanna unify Expr and GenExpr.
# Maybe when consexpr is released it makes sense to revisit this.
# TODO: We have to think about the operations that we define: __isub__, __add__, etc
# and when to copy expressions and when to not copy them.
# For example: when creating a ExprCons from an Expr expr, we store the expression expr
# and then we normalize. When doing the normalization, we do
# ```
# c = self.expr[CONST]
# self.expr -= c
# ```
# which should, in princple, modify the expr. However, since we do not implement __isub__, __sub__
# gets called (I guess) and so a copy is returned.
# Modifying the expression directly would be a bug, given that the expression might be re-used by the user. </pre>
import math
from typing import TYPE_CHECKING, Literal, Union

import numpy as np

from cpython.dict cimport PyDict_Next, PyDict_GetItem
from cpython.float cimport PyFloat_Check
from cpython.long cimport PyLong_Check
from cpython.number cimport PyNumber_Check
from cpython.object cimport Py_LE, Py_EQ, Py_GE, Py_TYPE
from cpython.ref cimport PyObject
from cpython.tuple cimport PyTuple_GET_ITEM

from pyscipopt.scip cimport Variable, Solution


if TYPE_CHECKING:
    double = float


cdef class Term:
    '''This is a monomial term'''

    cdef readonly tuple vartuple
    cdef Py_ssize_t hashval

    def __init__(self, *vartuple: Variable):
        self.vartuple = tuple(sorted(vartuple, key=lambda v: v.getIndex()))
        self.hashval = <Py_ssize_t>hash(tuple(v.ptr() for v in self.vartuple))

    def __getitem__(self, idx):
        return self.vartuple[idx]

    def __hash__(self) -> Py_ssize_t:
        return self.hashval

    def __eq__(self, other) -> bool:
        if other is self:
            return True
        if <type>Py_TYPE(other) is not Term:
            return False

        cdef int n = len(self)
        cdef Term _other = <Term>other
        if n != len(_other) or self.hashval != _other.hashval:
            return False

        cdef int i
        cdef Variable var1, var2
        for i in range(n):
            var1 = <Variable>PyTuple_GET_ITEM(self.vartuple, i)
            var2 = <Variable>PyTuple_GET_ITEM(_other.vartuple, i)
            if var1.ptr() != var2.ptr():
                return False
        return True

    def __len__(self):
        return len(self.vartuple)

    def __mul__(self, Term other):
        # NOTE: This merge algorithm requires a sorted `Term.vartuple`.
        # This should be ensured in the constructor of Term.
        cdef int n1 = len(self)
        cdef int n2 = len(other)
        if n1 == 0: return other
        if n2 == 0: return self

        cdef list vartuple = [None] * (n1 + n2)
        cdef int i = 0, j = 0, k = 0
        cdef Variable var1, var2
        while i < n1 and j < n2:
            var1 = <Variable>PyTuple_GET_ITEM(self.vartuple, i)
            var2 = <Variable>PyTuple_GET_ITEM(other.vartuple, j)
            if var1.getIndex() <= var2.getIndex():
                vartuple[k] = var1
                i += 1
            else:
                vartuple[k] = var2
                j += 1
            k += 1
        while i < n1:
            vartuple[k] = <Variable>PyTuple_GET_ITEM(self.vartuple, i)
            i += 1
            k += 1
        while j < n2:
            vartuple[k] = <Variable>PyTuple_GET_ITEM(other.vartuple, j)
            j += 1
            k += 1

        cdef Term res = Term.__new__(Term)
        res.vartuple = tuple(vartuple)
        res.hashval = <Py_ssize_t>hash(tuple(v.ptr() for v in res.vartuple))
        return res

    def __repr__(self):
        return 'Term(%s)' % ', '.join([str(v) for v in self.vartuple])

    cpdef double _evaluate(self, Solution sol) except *:
        cdef double res = 1.0
        cdef SCIP* scip_ptr = sol.scip
        cdef SCIP_SOL* sol_ptr = sol.sol
        cdef int i = 0, n = len(self)
        cdef Variable var

        for i in range(n):
            var = <Variable>self.vartuple[i]
            res *= SCIPgetSolVal(scip_ptr, sol_ptr, var.scip_var)
            if res == 0:  # early stop
                return 0.0
        return res


CONST = Term()


# helper function
def buildGenExprObj(expr: Union[int, float, np.number, Expr, GenExpr]) -> GenExpr:
    """helper function to generate an object of type GenExpr"""
    if not _is_genexpr_compatible(expr):
        raise TypeError(f"unsupported type {type(expr).__name__!s}")

    if _is_number(expr):
        return Constant(expr)

    elif isinstance(expr, Expr):
        # loop over terms and create a sumexpr with the sum of each term
        # each term is either a variable (which gets transformed into varexpr)
        # or a product of variables (which gets tranformed into a prod)
        sumexpr = SumExpr()
        for vars, coef in expr.terms.items():
            if len(vars) == 0:
                sumexpr += coef
            elif len(vars) == 1:
                varexpr = VarExpr(vars[0])
                sumexpr += coef * varexpr
            else:
                prodexpr = ProdExpr()
                for v in vars:
                    varexpr = VarExpr(v)
                    prodexpr *= varexpr
                sumexpr += coef * prodexpr
        return sumexpr

    return expr


cdef class ExprLike:

    def __array_ufunc__(
        self,
        ufunc: np.ufunc,
        method: Literal["__call__", "reduce", "reduceat", "accumulate", "outer", "at"],
        *args,
        **kwargs,
    ):
        if kwargs.get("out", None) is not None:
            raise TypeError(
                f"{self.__class__.__name__} doesn't support the 'out' parameter in __array_ufunc__"
            )

        if method == "__call__":
            if arrays := [a for a in args if isinstance(a, np.ndarray)]:
                if any(a.dtype.kind not in "fiub" for a in arrays):
                    return NotImplemented
                # If the np.ndarray is of numeric type, all arguments are converted to
                # MatrixExpr or MatrixGenExpr and then the ufunc is applied.
                return ufunc(*[_ensure_matrix(a) for a in args], **kwargs)

            # Convert `np.generic` to native Python types to stop __array_ufunc__
            # recursion from `np.generic + MatrixExpr`.
            args = [a.item() if isinstance(a, np.generic) else a for a in args]

            if ufunc is np.add:
                return args[0] + args[1]
            elif ufunc is np.subtract:
                return args[0] - args[1]
            elif ufunc is np.multiply:
                return args[0] * args[1]
            elif ufunc in {np.divide, np.true_divide}:
                return args[0] / args[1]
            elif ufunc is np.power:
                return args[0] ** args[1]
            elif ufunc is np.negative:
                return -args[0]
            elif ufunc is np.less_equal:
                return args[0] <= args[1]
            elif ufunc is np.greater_equal:
                return args[0] >= args[1]
            elif ufunc is np.equal:
                return args[0] == args[1]
            elif ufunc is np.absolute:
                return args[0].__abs__()
            elif ufunc is np.exp:
                return args[0].exp()
            elif ufunc is np.log:
                return args[0].log()
            elif ufunc is np.sqrt:
                return args[0].sqrt()
            elif ufunc is np.sin:
                return args[0].sin()
            elif ufunc is np.cos:
                return args[0].cos()

        return NotImplemented

    def __abs__(self) -> GenExpr:
        return UnaryExpr(Operator.fabs, buildGenExprObj(self))

    def exp(self) -> GenExpr:
        return UnaryExpr(Operator.exp, buildGenExprObj(self))

    def log(self) -> GenExpr:
        return UnaryExpr(Operator.log, buildGenExprObj(self))

    def sqrt(self) -> GenExpr:
        return UnaryExpr(Operator.sqrt, buildGenExprObj(self))

    def sin(self) -> GenExpr:
        return UnaryExpr(Operator.sin, buildGenExprObj(self))

    def cos(self) -> GenExpr:
        return UnaryExpr(Operator.cos, buildGenExprObj(self))


##@details Polynomial expressions of variables with operator overloading. \n
#See also the @ref ExprDetails "description" in the expr.pxi. 
cdef class Expr(ExprLike):

    def __init__(self, terms=None):
        '''terms is a dict of variables to coefficients.

        CONST is used as key for the constant term.'''
        self.terms = {} if terms is None else terms

        if len(self.terms) == 0:
            self.terms[CONST] = 0.0

    def __getitem__(self, key):
        if not isinstance(key, Term):
            key = Term(key)
        return self.terms.get(key, 0.0)

    def __iter__(self):
        return iter(self.terms)

    def __add__(self, other):
        if not _is_expr_compatible(other):
            return NotImplemented

        left = self
        right = other
        terms = left.terms.copy()

        if isinstance(right, Expr):
            # merge the terms by component-wise addition
            for v,c in right.terms.items():
                terms[v] = terms.get(v, 0.0) + c
        elif _is_number(right):
            c = float(right)
            terms[CONST] = terms.get(CONST, 0.0) + c
        return Expr(terms)

    def __iadd__(self, other):
        if not _is_expr_compatible(other):
            return NotImplemented

        if isinstance(other, Expr):
            for v,c in other.terms.items():
                self.terms[v] = self.terms.get(v, 0.0) + c
        elif _is_number(other):
            c = float(other)
            self.terms[CONST] = self.terms.get(CONST, 0.0) + c
        return self

    def __mul__(self, other):
        if not _is_expr_compatible(other):
            return NotImplemented

        cdef dict res = {}
        cdef Py_ssize_t pos1 = <Py_ssize_t>0, pos2 = <Py_ssize_t>0
        cdef PyObject *k1_ptr = NULL
        cdef PyObject *v1_ptr = NULL
        cdef PyObject *k2_ptr = NULL
        cdef PyObject *v2_ptr = NULL
        cdef PyObject *old_v_ptr = NULL
        cdef Term child
        cdef double coef

        if _is_number(other):
            coef = <double>other
            while PyDict_Next(self.terms, &pos1, &k1_ptr, &v1_ptr):
                res[<Term>k1_ptr] = <double>(<object>v1_ptr) * coef

        elif isinstance(other, Expr):
            while PyDict_Next(self.terms, &pos1, &k1_ptr, &v1_ptr):
                pos2 = <Py_ssize_t>0
                while PyDict_Next(other.terms, &pos2, &k2_ptr, &v2_ptr):
                    child = (<Term>k1_ptr) * (<Term>k2_ptr)
                    coef = (<double>(<object>v1_ptr)) * (<double>(<object>v2_ptr))
                    if (old_v_ptr := PyDict_GetItem(res, child)) != NULL:
                        res[child] = <double>(<object>old_v_ptr) + coef
                    else:
                        res[child] = coef
        return Expr(res)

    def __truediv__(self, other):
        if not _is_expr_compatible(other):
            return NotImplemented

        if _is_number(other):
            return 1.0 / other * self
        return buildGenExprObj(self) / other

    def __rtruediv__(self, other):
        ''' other / self '''
        if not _is_expr_compatible(other):
            return NotImplemented
        return buildGenExprObj(other) / self

    def __pow__(self, other, modulo):
        if float(other).is_integer() and other >= 0:
            exp = int(other)
        else: # need to transform to GenExpr
            return buildGenExprObj(self)**other

        res = 1
        for _ in range(exp):
            res *= self
        return res

    def __rpow__(self, other):
        """
        Implements base**x as scip.exp(x * scip.log(base)).
        Note: base must be positive.
        """
        if not _is_number(other):
            raise TypeError(f"Unsupported base type {type(other)} for exponentiation.")

        if (base := <double>other) <= 0.0:
            raise ValueError("Base of a**x must be positive, as expression is reformulated to scip.exp(x * scip.log(a)); got %g" % base)
        return (self * Constant(base).log()).exp()

    def __neg__(self):
        return Expr({v:-c for v,c in self.terms.items()})

    def __sub__(self, other):
        return self + (-other)

    def __radd__(self, other):
        return self.__add__(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __rsub__(self, other):
        return -1.0 * self + other

    def __richcmp__(self, other, int op):
        '''turn it into a constraint'''
        return _expr_richcmp(self, other, op)

    def normalize(self):
        '''remove terms with coefficient of 0'''
        self.terms =  {t:c for (t,c) in self.terms.items() if c != 0.0}

    def __repr__(self):
        return 'Expr(%s)' % repr(self.terms)

    def degree(self):
        '''computes highest degree of terms'''
        if len(self.terms) == 0:
            return 0
        else:
            return max(len(v) for v in self.terms)

    cpdef double _evaluate(self, Solution sol) except *:
        cdef double res = 0
        cdef Py_ssize_t pos = <Py_ssize_t>0
        cdef PyObject* key_ptr
        cdef PyObject* val_ptr
        cdef Term term
        cdef double coef

        while PyDict_Next(self.terms, &pos, &key_ptr, &val_ptr):
            term = <Term>key_ptr
            coef = <double>(<object>val_ptr)
            res += coef * term._evaluate(sol)
        return res


cdef class ExprCons:
    '''Constraints with a polynomial expressions and lower/upper bounds.'''
    cdef public expr
    cdef public _lhs
    cdef public _rhs

    def __init__(self, expr, lhs=None, rhs=None):
        self.expr = expr
        self._lhs = lhs
        self._rhs = rhs
        assert not (lhs is None and rhs is None)
        self.normalize()

    def normalize(self):
        '''move constant terms in expression to bounds'''
        if isinstance(self.expr, Expr):
            c = self.expr[CONST]
            self.expr -= c
            assert self.expr[CONST] == 0.0
            self.expr.normalize()
        else:
            assert isinstance(self.expr, GenExpr)
            return

        if not self._lhs is None:
            self._lhs -= c
        if not self._rhs is None:
            self._rhs -= c


    def __richcmp__(self, other, op):
        '''turn it into a constraint'''
        if not _is_number(other):
            raise TypeError('Ranged ExprCons is not well defined!')

        if op == 1: # <=
            if not self._rhs is None:
                raise TypeError('ExprCons already has upper bound')
            assert not self._lhs is None

            return ExprCons(self.expr, lhs=self._lhs, rhs=float(other))
        elif op == 5: # >=
            if not self._lhs is None:
                raise TypeError('ExprCons already has lower bound')
            assert self._lhs is None
            assert not self._rhs is None

            return ExprCons(self.expr, lhs=float(other), rhs=self._rhs)
        else:
            raise NotImplementedError("Ranged ExprCons can only support with '<=' or '>='.")

    def __repr__(self):
        return 'ExprCons(%s, %s, %s)' % (self.expr, self._lhs, self._rhs)

    def __bool__(self):
        '''Make sure that equality of expressions is not asserted with =='''

        msg = """Can't evaluate constraints as booleans.

If you want to add a ranged constraint of the form
   lhs <= expression <= rhs
you have to use parenthesis to break the Python syntax for chained comparisons:
   lhs <= (expression <= rhs)
"""
        raise TypeError(msg)

def quicksum(termlist):
    '''add linear expressions and constants much faster than Python's sum
    by avoiding intermediate data structures and adding terms inplace
    '''
    result = Expr()
    for term in termlist:
        result += term
    return result

def quickprod(termlist):
    '''multiply linear expressions and constants by avoiding intermediate 
    data structures and multiplying terms inplace
    '''
    result = Expr() + 1
    for term in termlist:
        result *= term
    return result


class Op:
    const = 'const'
    varidx = 'var'
    exp, log, sqrt, sin, cos = 'exp', 'log', 'sqrt', 'sin', 'cos'
    plus, minus, mul, div, power = '+', '-', '*', '/', '**'
    add = 'sum'
    prod = 'prod'
    fabs = 'abs'

Operator = Op()

##@details <pre> General expressions of variables with operator overloading.
#
#@note
#   - these expressions are not smart enough to identify equal terms
#   - in contrast to polynomial expressions, __getitem__ is not implemented
#     so expr[x] will generate an error instead of returning the coefficient of x </pre>
#
#See also the @ref ExprDetails "description" in the expr.pxi. 
cdef class GenExpr(ExprLike):

    cdef public _op
    cdef public children

    def __init__(self): # do we need it
        ''' '''

    def __add__(self, other):
        if not _is_genexpr_compatible(other):
            return NotImplemented

        left = buildGenExprObj(self)
        right = buildGenExprObj(other)
        ans = SumExpr()

        # add left term
        if left.getOp() == Operator.add:
            ans.coefs.extend(left.coefs)
            ans.children.extend(left.children)
            ans.constant += left.constant
        elif left.getOp() == Operator.const:
            ans.constant += left.number
        else:
            ans.coefs.append(1.0)
            ans.children.append(left)

        # add right term
        if right.getOp() == Operator.add:
            ans.coefs.extend(right.coefs)
            ans.children.extend(right.children)
            ans.constant += right.constant
        elif right.getOp() == Operator.const:
            ans.constant += right.number
        else:
            ans.coefs.append(1.0)
            ans.children.append(right)

        return ans

    #def __iadd__(self, other):
    #''' in-place addition, i.e., expr += other '''
    #    assert isinstance(self, Expr)
    #    right = buildGenExprObj(other)
    #
    #    # transform self into sum
    #    if self.getOp() != Operator.add:
    #        newsum = SumExpr()
    #        if self.getOp() == Operator.const:
    #            newsum.constant += self.number
    #        else:
    #            newsum.coefs.append(1.0)
    #            newsum.children.append(self.copy()) # TODO: what is copy?
    #        self = newsum
    #    # add right term
    #    if right.getOp() == Operator.add:
    #        self.coefs.extend(right.coefs)
    #        self.children.extend(right.children)
    #        self.constant += right.constant
    #    elif right.getOp() == Operator.const:
    #        self.constant += right.number
    #    else:
    #        self.coefs.append(1.0)
    #        self.children.append(right)
    #    return self

    def __mul__(self, other):
        if not _is_genexpr_compatible(other):
            return NotImplemented

        left = buildGenExprObj(self)
        right = buildGenExprObj(other)
        ans = ProdExpr()

        # multiply left factor
        if left.getOp() == Operator.prod:
            ans.children.extend(left.children)
            ans.constant *= left.constant
        elif left.getOp() == Operator.const:
            ans.constant *= left.number
        else:
            ans.children.append(left)

        # multiply right factor
        if right.getOp() == Operator.prod:
            ans.children.extend(right.children)
            ans.constant *= right.constant
        elif right.getOp() == Operator.const:
            ans.constant *= right.number
        else:
            ans.children.append(right)

        return ans

    #def __imul__(self, other):
    #''' in-place multiplication, i.e., expr *= other '''
    #    assert isinstance(self, Expr)
    #    right = buildGenExprObj(other)
    #    # transform self into prod
    #    if self.getOp() != Operator.prod:
    #        newprod = ProdExpr()
    #        if self.getOp() == Operator.const:
    #            newprod.constant *= self.number
    #        else:
    #            newprod.children.append(self.copy()) # TODO: what is copy?
    #        self = newprod
    #    # multiply right factor
    #    if right.getOp() == Operator.prod:
    #        self.children.extend(right.children)
    #        self.constant *= right.constant
    #    elif right.getOp() == Operator.const:
    #        self.constant *= right.number
    #    else:
    #        self.children.append(right)
    #    return self

    def __pow__(self, other, modulo):
        expo = buildGenExprObj(other)
        if expo.getOp() != Operator.const:
            raise NotImplementedError("exponents must be numbers")
        if self.getOp() == Operator.const:
            return Constant(self.number**expo.number)
        ans = PowExpr()
        ans.children.append(self)
        ans.expo = expo.number

        return ans

    def __rpow__(self, other):
        """
        Implements base**x as scip.exp(x * scip.log(base)). 
        Note: base must be positive.
        """
        if not _is_number(other):
            raise TypeError(f"Unsupported base type {type(other)} for exponentiation.")

        if (base := <double>other) <= 0.0:
            raise ValueError("Base of a**x must be positive, as expression is reformulated to scip.exp(x * scip.log(a)); got %g" % base)
        return (self * Constant(base).log()).exp()

    #TODO: ipow, idiv, etc
    def __truediv__(self,other):
        if not _is_genexpr_compatible(other):
            return NotImplemented

        divisor = buildGenExprObj(other)
        # we can't divide by 0
        if isinstance(divisor, GenExpr) and divisor.getOp() == Operator.const and divisor.number == 0.0:
            raise ZeroDivisionError("cannot divide by 0")
        return self * divisor**(-1)

    def __rtruediv__(self, other):
        ''' other / self '''
        if not _is_genexpr_compatible(other):
            return NotImplemented
        return buildGenExprObj(other) / self

    def __neg__(self):
        return -1.0 * self

    def __sub__(self, other):
        return self + (-other)

    def __radd__(self, other):
        return self.__add__(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __rsub__(self, other):
        return -1.0 * self + other

    def __richcmp__(self, other, int op):
        '''turn it into a constraint'''
        return _expr_richcmp(self, other, op)

    def degree(self):
        '''Note: none of these expressions should be polynomial'''
        return float('inf') 

    def getOp(self):
        '''returns operator of GenExpr'''
        return self._op

    cdef GenExpr copy(self, bool copy = True):
        cdef object cls = <type>Py_TYPE(self)
        cdef GenExpr res = cls.__new__(cls)
        res._op = self._op
        res.children = self.children.copy() if copy else self.children
        if cls is SumExpr:
            (<SumExpr>res).constant = (<SumExpr>self).constant
            (<SumExpr>res).coefs = (<SumExpr>self).coefs.copy() if copy else (<SumExpr>self).coefs
        if cls is ProdExpr:
            (<ProdExpr>res).constant = (<ProdExpr>self).constant
        elif cls is PowExpr:
            (<PowExpr>res).expo = (<PowExpr>self).expo
        return res


# Sum Expressions
cdef class SumExpr(GenExpr):

    cdef public constant
    cdef public coefs

    def __init__(self):
        self.constant = 0.0
        self.coefs = []
        self.children = []
        self._op = Operator.add
    def __repr__(self):
        return self._op + "(" + str(self.constant) + "," + ",".join(map(lambda child : child.__repr__(), self.children)) + ")"

    cpdef double _evaluate(self, Solution sol) except *:
        cdef double res = self.constant
        cdef int i = 0, n = len(self.children)
        cdef list children = self.children
        cdef list coefs = self.coefs
        for i in range(n):
            res += <double>coefs[i] * (<GenExpr>children[i])._evaluate(sol)
        return res


# Prod Expressions
cdef class ProdExpr(GenExpr):

    cdef public constant

    def __init__(self):
        self.constant = 1.0
        self.children = []
        self._op = Operator.prod

    def __repr__(self):
        return self._op + "(" + str(self.constant) + "," + ",".join(map(lambda child : child.__repr__(), self.children)) + ")"

    cpdef double _evaluate(self, Solution sol) except *:
        cdef double res = self.constant
        cdef list children = self.children
        cdef int i = 0, n = len(children)
        for i in range(n):
            res *= (<GenExpr>children[i])._evaluate(sol)
            if res == 0:  # early stop
                return 0.0
        return res


# Var Expressions
cdef class VarExpr(GenExpr):

    cdef public var

    def __init__(self, var):
        self.children = [var]
        self._op = Operator.varidx

    def __repr__(self):
        return self.children[0].__repr__()

    cpdef double _evaluate(self, Solution sol) except *:
        return (<Expr>self.children[0])._evaluate(sol)


# Pow Expressions
cdef class PowExpr(GenExpr):

    cdef public expo

    def __init__(self):
        self.expo = 1.0
        self.children = []
        self._op = Operator.power

    def __repr__(self):
        return self._op + "(" + self.children[0].__repr__() + "," + str(self.expo) + ")"

    cpdef double _evaluate(self, Solution sol) except *:
        return (<GenExpr>self.children[0])._evaluate(sol) ** self.expo


# Exp, Log, Sqrt, Sin, Cos Expressions
cdef class UnaryExpr(GenExpr):
    def __init__(self, op, expr):
        self.children = []
        self.children.append(expr)
        self._op = op

    def __abs__(self) -> UnaryExpr:
        if self._op == "abs":
            return <UnaryExpr>self.copy()
        return UnaryExpr(Operator.fabs, self)

    def __repr__(self):
        return self._op + "(" + self.children[0].__repr__() + ")"

    cpdef double _evaluate(self, Solution sol) except *:
        cdef double res = (<GenExpr>self.children[0])._evaluate(sol)
        return math.fabs(res) if self._op == "abs" else getattr(math, self._op)(res)


# class for constant expressions
cdef class Constant(GenExpr):
    cdef public number
    def __init__(self,number):
        self.number = number
        self._op = Operator.const

    def __repr__(self):
        return str(self.number)

    cpdef double _evaluate(self, Solution sol) except *:
        return self.number


def exp(x):
    """
    returns expression with exp-function

    Parameters
    ----------
    x : Expr, GenExpr, number, np.ndarray, list, or tuple
        - If x is a scalar expression or number, apply the exp function directly to it.
          And if it's a number, convert it to a Constant expression first.
        - If x is a vector (np.ndarray, list, or tuple), apply the exp function
          element-wise using np.frompyfunc to convert each element to a Constant if it's
          a number, and then apply the exp function.

    Returns
    -------
    GenExpr or MatrixGenExpr
        - If x is a scalar expression or number, returns the result of applying the exp
          function to it.
        - If x is a vector, returns an np.ndarray of the same shape with the exp
          function applied element-wise.
    """
    return _wrap_ufunc(x, np.exp)


def log(x):
    """
    returns expression with log-function

    Parameters
    ----------
    x : Expr, GenExpr, number, np.ndarray, list, or tuple
        - If x is a scalar expression or number, apply the log function directly to it.
          And if it's a number, convert it to a Constant expression first.
        - If x is a vector (np.ndarray, list, or tuple), apply the log function
          element-wise using np.frompyfunc to convert each element to a Constant if it's
          a number, and then apply the log function.

    Returns
    -------
    GenExpr or MatrixGenExpr
        - If x is a scalar expression or number, returns the result of applying the log
          function to it.
        - If x is a vector, returns an np.ndarray of the same shape with the log
          function applied element-wise.
    """
    return _wrap_ufunc(x, np.log)


def sqrt(x):
    """
    returns expression with sqrt-function

    Parameters
    ----------
    x : Expr, GenExpr, number, np.ndarray, list, or tuple
        - If x is a scalar expression or number, apply the sqrt function directly to it.
          And if it's a number, convert it to a Constant expression first.
        - If x is a vector (np.ndarray, list, or tuple), apply the sqrt function
          element-wise using np.frompyfunc to convert each element to a Constant if it's
          a number, and then apply the sqrt function.

    Returns
    -------
    GenExpr or MatrixGenExpr
        - If x is a scalar expression or number, returns the result of applying the sqrt
          function to it.
        - If x is a vector, returns an np.ndarray of the same shape with the sqrt
          function applied element-wise.
    """
    return _wrap_ufunc(x, np.sqrt)


def sin(x):
    """
    returns expression with sin-function

    Parameters
    ----------
    x : Expr, GenExpr, number, np.ndarray, list, or tuple
        - If x is a scalar expression or number, apply the sin function directly to it.
          And if it's a number, convert it to a Constant expression first.
        - If x is a vector (np.ndarray, list, or tuple), apply the sin function
          element-wise using np.frompyfunc to convert each element to a Constant if it's
          a number, and then apply the sin function.

    Returns
    -------
    GenExpr or MatrixGenExpr
        - If x is a scalar expression or number, returns the result of applying the sin
          function to it.
        - If x is a vector, returns an np.ndarray of the same shape with the sin
          function applied element-wise.
    """
    return _wrap_ufunc(x, np.sin)


def cos(x):
    """
    returns expression with cos-function

    Parameters
    ----------
    x : Expr, GenExpr, number, np.ndarray, list, or tuple
        - If x is a scalar expression or number, apply the cos function directly to it.
          And if it's a number, convert it to a Constant expression first.
        - If x is a vector (np.ndarray, list, or tuple), apply the cos function
          element-wise using np.frompyfunc to convert each element to a Constant if it's
          a number, and then apply the cos function.

    Returns
    -------
    GenExpr or MatrixGenExpr
        - If x is a scalar expression or number, returns the result of applying the cos
          function to it.
        - If x is a vector, returns an np.ndarray of the same shape with the cos
          function applied element-wise.
    """
    return _wrap_ufunc(x, np.cos)


cdef inline object _to_const(object x):
    return Constant(<double>x) if _is_number(x) else x

cdef object _vec_to_const = np.frompyfunc(_to_const, 1, 1)

cdef inline object _wrap_ufunc(object x, object ufunc):
    """
    Apply a universal function (ufunc) to an expression or a collection of expressions.

    Parameters
    ----------
    x : Expr, GenExpr, number, np.ndarray, list, or tuple
        - If x is a scalar expression or number, apply the ufunc directly to it. And if
          it's a number, convert it to a Constant expression first.
        - If x is a vector (np.ndarray, list, or tuple), apply the ufunc element-wise
          using np.frompyfunc to convert each element to a Constant if it's a number,
          and then apply the ufunc.

    ufunc : np.ufunc
        The universal function to be applied to x.

    Returns
    -------
    GenExpr or MatrixGenExpr
        - If x is a scalar expression or number, returns the result of applying the
          ufunc to it.
        - If x is a vector, returns an np.ndarray of the same shape with the ufunc
          applied element-wise.
    """
    if isinstance(x, (np.ndarray, list, tuple)):
        res = ufunc(_vec_to_const(x))
        return res.view(MatrixGenExpr) if isinstance(res, np.ndarray) else res
    return ufunc(_to_const(x))

cdef inline object _ensure_matrix(object arg):
    if type(arg) is np.ndarray:
        return arg.view(MatrixExpr)
    matrix = MatrixExpr if isinstance(arg, Expr) else MatrixGenExpr
    return np.array(arg, dtype=object).view(matrix)


def expr_to_nodes(expr):
    '''transforms tree to an array of nodes. each node is an operator and the position of the 
    children of that operator (i.e. the other nodes) in the array'''
    assert isinstance(expr, GenExpr)
    nodes = []
    expr_to_array(expr, nodes)
    return nodes

def value_to_array(val, nodes):
    """adds a given value to an array"""
    nodes.append(tuple(['const', [val]]))
    return len(nodes) - 1

# there many hacky things here: value_to_array is trying to mimick
# the multiple dispatch of julia. Also that we have to ask which expression is which
# in order to get the constants correctly
# also, for sums, we are not considering coefficients, because basically all coefficients are 1
# haven't even consider substractions, but I guess we would interpret them as a - b = a + (-1) * b
def expr_to_array(expr, nodes):
    """adds expression to array"""
    op = expr._op
    if op == Operator.const: # FIXME: constant expr should also have children!
        nodes.append(tuple([op, [expr.number]]))
    elif op != Operator.varidx:
        indices = []
        nchildren = len(expr.children)
        for child in expr.children:
            pos = expr_to_array(child, nodes) # position of child in the final array of nodes, 'nodes'
            indices.append(pos)
        if op == Operator.power:
            pos = value_to_array(expr.expo, nodes)
            indices.append(pos)
        elif (op == Operator.add and expr.constant != 0.0) or (op == Operator.prod and expr.constant != 1.0):
            pos = value_to_array(expr.constant, nodes)
            indices.append(pos)
        nodes.append( tuple( [op, indices] ) )
    else: # var
        nodes.append( tuple( [op, expr.children] ) )
    return len(nodes) - 1


cdef inline bint _is_number(object x):
    if PyLong_Check(x) or PyFloat_Check(x):
        return True
    if cnp.PyArray_Check(x) or isinstance(x, (ExprLike, list, tuple)):
        return False
    return PyNumber_Check(x)

cdef inline bint _is_expr_compatible(object x):
    return _is_number(x) or isinstance(x, Expr)

cdef inline bint _is_genexpr_compatible(object x):
    return _is_expr_compatible(x) or isinstance(x, GenExpr)

cdef object _expr_richcmp(
    ExprLike self,
    other: Union[int, float, np.number, Expr, GenExpr],
    int op,
):
    if isinstance(other, np.ndarray):
        return NotImplemented
    if not _is_genexpr_compatible(other):
        raise TypeError(f"unsupported type {type(other).__name__!s}")

    if op == Py_LE:
        if _is_number(other):
            return ExprCons(self, rhs=<double>other)
        return ExprCons(self - other, rhs=0.0)
    elif op == Py_GE:
        if _is_number(other):
            return ExprCons(self, lhs=<double>other)
        return ExprCons(self - other, lhs=0.0)
    elif op == Py_EQ:
        if _is_number(other):
            return ExprCons(self, lhs=<double>other, rhs=<double>other)
        return ExprCons(self - other, lhs=0.0, rhs=0.0)
    raise NotImplementedError("can only support with '<=', '>=', or '=='")



from pyscipopt import Model, Constraint
from typing import List

def getLocalConss(model: Model, node = None) -> List[List[Constraint]]:
    """
    Returns local constraints.

    Parameters
    ----------
    model : Model
        The model from which to retrieve the local constraints.
    node : Node, optional
        The node from which to retrieve the local constraints. If not provided, the current node is used.

    Returns
    -------
    list[Constraint]
        A list of local constraints. First entry are global constraints, second entry are all the added constraints.
    """

    if node is None:
        assert model.getStageName() in ["INITPRESOLVE", "PRESOLVING", "EXITPRESOLVE", "SOLVING"], "Model cannot be called in stage %s." % model.getStageName()
        cur_node = model.getCurrentNode()
    else:
        cur_node = node

    added_conss: List[Constraint] = []
    while cur_node is not None:
        added_conss = cur_node.getAddedConss() + added_conss
        cur_node = cur_node.getParent()
    
    return [model.getConss(), added_conss]

def getNLocalConss(model: Model, node = None) -> List[int]:
    """
    Returns the number of local constraints of a node.

    Parameters
    ----------
    model : Model
        The model from which to retrieve the number of local constraints.
    node : Node, optional
        The node from which to retrieve the number of local constraints. If not provided, the current node is used.

    Returns
    -------
    list[int]
        A list of the number of local constraints. First entry is the number of global constraints, second entry is the number of all the added constraints.
    """
    local_conss = getLocalConss(model, node)
    return [len(local_conss[0]), len(local_conss[1])]


##@file benders.pxi
#@brief Base class of the Benders decomposition Plugin
cdef class Benders:
    cdef public Model model
    cdef public str name
    cdef SCIP_BENDERS* _benders

    def bendersfree(self):
        '''calls destructor and frees memory of Benders decomposition '''
        pass

    def bendersinit(self):
        '''initializes Benders deconposition'''
        pass

    def bendersexit(self):
        '''calls exit method of Benders decomposition'''
        pass

    def bendersinitpre(self):
        '''informs the Benders decomposition that the presolving process is being started '''
        pass

    def bendersexitpre(self):
        '''informs the Benders decomposition that the presolving process has been completed'''
        pass

    def bendersinitsol(self):
        '''informs Benders decomposition that the branch and bound process is being started '''
        pass

    def bendersexitsol(self):
        '''informs Benders decomposition that the branch and bound process data is being freed'''
        pass

    def benderscreatesub(self, probnumber):
        '''creates the subproblems and registers it with the Benders decomposition struct '''
        raise NotImplementedError("benderscreatesub() is a fundamental callback and should be implemented in the derived class")

    def benderspresubsolve(self, solution, enfotype, checkint):
        '''sets the pre subproblem solve callback of Benders decomposition '''
        return {}

    def benderssolvesubconvex(self, solution, probnumber, onlyconvex):
        '''sets convex solve callback of Benders decomposition'''
        return {}

    def benderssolvesub(self, solution, probnumber):
        '''sets solve callback of Benders decomposition '''
        return {}

    def benderspostsolve(self, solution, enfotype, mergecandidates, npriomergecands, checkint, infeasible):
        '''sets post-solve callback of Benders decomposition '''
        return {}

    def bendersfreesub(self, probnumber):
        '''frees the subproblems'''
        pass

    def bendersgetvar(self, variable, probnumber):
        '''Returns the corresponding master or subproblem variable for the given variable. This provides a call back for the variable mapping between the master and subproblems. '''
        raise NotImplementedError("bendersgetvar() is a fundamental callback and should be implemented in the derived class")

# local helper functions for the interface
cdef Variable getPyVar(SCIP_VAR* var):
    cdef SCIP_VARDATA* vardata
    vardata = SCIPvarGetData(var)
    return <Variable>vardata


cdef SCIP_RETCODE PyBendersCopy (SCIP* scip, SCIP_BENDERS* benders, SCIP_Bool threadsafe) noexcept with gil:
    return SCIP_OKAY

cdef SCIP_RETCODE PyBendersFree (SCIP* scip, SCIP_BENDERS* benders) noexcept with gil:
    cdef SCIP_BENDERSDATA* bendersdata
    bendersdata = SCIPbendersGetData(benders)
    PyBenders = <Benders>bendersdata
    PyBenders.bendersfree()
    return SCIP_OKAY

cdef SCIP_RETCODE PyBendersInit (SCIP* scip, SCIP_BENDERS* benders) noexcept with gil:
    cdef SCIP_BENDERSDATA* bendersdata
    bendersdata = SCIPbendersGetData(benders)
    PyBenders = <Benders>bendersdata
    PyBenders.bendersinit()
    return SCIP_OKAY

cdef SCIP_RETCODE PyBendersExit (SCIP* scip, SCIP_BENDERS* benders) noexcept with gil:
    cdef SCIP_BENDERSDATA* bendersdata
    bendersdata = SCIPbendersGetData(benders)
    PyBenders = <Benders>bendersdata
    PyBenders.bendersexit()
    return SCIP_OKAY

cdef SCIP_RETCODE PyBendersInitpre (SCIP* scip, SCIP_BENDERS* benders) noexcept with gil:
    cdef SCIP_BENDERSDATA* bendersdata
    bendersdata = SCIPbendersGetData(benders)
    PyBenders = <Benders>bendersdata
    PyBenders.bendersinitpre()
    return SCIP_OKAY

cdef SCIP_RETCODE PyBendersExitpre (SCIP* scip, SCIP_BENDERS* benders) noexcept with gil:
    cdef SCIP_BENDERSDATA* bendersdata
    bendersdata = SCIPbendersGetData(benders)
    PyBenders = <Benders>bendersdata
    PyBenders.bendersexitpre()
    return SCIP_OKAY

cdef SCIP_RETCODE PyBendersInitsol (SCIP* scip, SCIP_BENDERS* benders) noexcept with gil:
    cdef SCIP_BENDERSDATA* bendersdata
    bendersdata = SCIPbendersGetData(benders)
    PyBenders = <Benders>bendersdata
    PyBenders.bendersinitsol()
    return SCIP_OKAY

cdef SCIP_RETCODE PyBendersExitsol (SCIP* scip, SCIP_BENDERS* benders) noexcept with gil:
    cdef SCIP_BENDERSDATA* bendersdata
    bendersdata = SCIPbendersGetData(benders)
    PyBenders = <Benders>bendersdata
    PyBenders.bendersexitsol()
    return SCIP_OKAY

cdef SCIP_RETCODE PyBendersCreatesub (SCIP* scip, SCIP_BENDERS* benders, int probnumber) noexcept with gil:
    cdef SCIP_BENDERSDATA* bendersdata
    bendersdata = SCIPbendersGetData(benders)
    PyBenders = <Benders>bendersdata
    PyBenders.benderscreatesub(probnumber)
    return SCIP_OKAY

cdef SCIP_RETCODE PyBendersPresubsolve (SCIP* scip, SCIP_BENDERS* benders, SCIP_SOL* sol, SCIP_BENDERSENFOTYPE type, SCIP_Bool checkint, SCIP_Bool* infeasible, SCIP_Bool* auxviol, SCIP_Bool* skipsolve,  SCIP_RESULT* result) noexcept with gil:
    cdef SCIP_BENDERSDATA* bendersdata
    bendersdata = SCIPbendersGetData(benders)
    PyBenders = <Benders>bendersdata
    if sol == NULL:
        solution = None
    else:
        solution = Solution.create(scip, sol)
    enfotype = type
    result_dict = PyBenders.benderspresubsolve(solution, enfotype, checkint)
    infeasible[0] = result_dict.get("infeasible", False)
    auxviol[0] = result_dict.get("auxviol", False)
    skipsolve[0] = result_dict.get("skipsolve", False)
    result[0] = result_dict.get("result", <SCIP_RESULT>result[0])
    return SCIP_OKAY

cdef SCIP_RETCODE PyBendersSolvesubconvex (SCIP* scip, SCIP_BENDERS* benders, SCIP_SOL* sol, int probnumber, SCIP_Bool onlyconvex, SCIP_Real* objective, SCIP_RESULT* result) noexcept with gil:
    cdef SCIP_BENDERSDATA* bendersdata
    bendersdata = SCIPbendersGetData(benders)
    PyBenders = <Benders>bendersdata
    if sol == NULL:
        solution = None
    else:
        solution = Solution.create(scip, sol)
    result_dict = PyBenders.benderssolvesubconvex(solution, probnumber, onlyconvex)
    objective[0] = result_dict.get("objective", 1e+20)
    result[0] = result_dict.get("result", <SCIP_RESULT>result[0])
    return SCIP_OKAY

cdef SCIP_RETCODE PyBendersSolvesub (SCIP* scip, SCIP_BENDERS* benders, SCIP_SOL* sol, int probnumber, SCIP_Real* objective, SCIP_RESULT* result) noexcept with gil:
    cdef SCIP_BENDERSDATA* bendersdata
    bendersdata = SCIPbendersGetData(benders)
    PyBenders = <Benders>bendersdata
    if sol == NULL:
        solution = None
    else:
        solution = Solution.create(scip, sol)
    result_dict = PyBenders.benderssolvesub(solution, probnumber)
    objective[0] = result_dict.get("objective", 1e+20)
    result[0] = result_dict.get("result", <SCIP_RESULT>result[0])
    return SCIP_OKAY

cdef SCIP_RETCODE PyBendersPostsolve (SCIP* scip, SCIP_BENDERS* benders, SCIP_SOL* sol,
        SCIP_BENDERSENFOTYPE type, int* mergecands, int npriomergecands, int nmergecands, SCIP_Bool checkint,
        SCIP_Bool infeasible, SCIP_Bool* merged) noexcept with gil:
    cdef SCIP_BENDERSDATA* bendersdata = SCIPbendersGetData(benders)
    cdef int i
    PyBenders = <Benders>bendersdata
    if sol == NULL:
        solution = None
    else:
        solution = Solution.create(scip, sol)
    enfotype = type
    mergecandidates = []
    for i in range(nmergecands):
        mergecandidates.append(mergecands[i])
    result_dict = PyBenders.benderspostsolve(solution, enfotype, mergecandidates, npriomergecands, checkint, infeasible)
    merged[0] = result_dict.get("merged", False)
    return SCIP_OKAY

cdef SCIP_RETCODE PyBendersFreesub (SCIP* scip, SCIP_BENDERS* benders, int probnumber) noexcept with gil:
    cdef SCIP_BENDERSDATA* bendersdata
    bendersdata = SCIPbendersGetData(benders)
    PyBenders = <Benders>bendersdata
    PyBenders.bendersfreesub(probnumber)
    return SCIP_OKAY

#TODO: Really need to ask about the passing and returning of variables
cdef SCIP_RETCODE PyBendersGetvar (SCIP* scip, SCIP_BENDERS* benders, SCIP_VAR* var, SCIP_VAR** mappedvar, int probnumber) noexcept with gil:
    cdef SCIP_BENDERSDATA* bendersdata
    bendersdata = SCIPbendersGetData(benders)
    PyBenders = <Benders>bendersdata
    PyVar = getPyVar(var)
    result_dict = PyBenders.bendersgetvar(PyVar, probnumber)
    mappedvariable = <Variable>(result_dict.get("mappedvar", None))
    if mappedvariable is None:
        mappedvar[0] = NULL
    else:
        mappedvar[0] = mappedvariable.scip_var
    return SCIP_OKAY



##@file lp.pxi
#@brief Base class of the LP Plugin
cdef class LP:
    cdef SCIP_LPI* lpi
    cdef readonly str name

    def __init__(self, name="LP", sense="minimize"):
        """
        Keyword arguments:
        name -- the name of the problem (default 'LP')
        sense -- objective sense (default minimize)
        """
        self.name = name
        n = str_conversion(name)
        if sense == "minimize":
            PY_SCIP_CALL(SCIPlpiCreate(&(self.lpi), NULL, n, SCIP_OBJSEN_MINIMIZE))
        elif sense == "maximize":
            PY_SCIP_CALL(SCIPlpiCreate(&(self.lpi), NULL, n, SCIP_OBJSEN_MAXIMIZE))
        else:
            raise Warning("unrecognized objective sense")

    def __dealloc__(self):
        PY_SCIP_CALL(SCIPlpiFree(&(self.lpi)))

    def __repr__(self):
        return self.name

    def writeLP(self, filename):
        """Writes LP to a file.

        Keyword arguments:
        filename -- the name of the file to be used
        """
        PY_SCIP_CALL(SCIPlpiWriteLP(self.lpi, filename))

    def readLP(self, filename):
        """Reads LP from a file.

        Keyword arguments:
        filename -- the name of the file to be used
        """
        PY_SCIP_CALL(SCIPlpiReadLP(self.lpi, filename))

    def infinity(self):
        """Returns infinity value of the LP.
        """
        return SCIPlpiInfinity(self.lpi)

    def isInfinity(self, val):
        """Checks if a given value is equal to the infinity value of the LP.

        Keyword arguments:
        val -- value that should be checked
        """
        return SCIPlpiIsInfinity(self.lpi, val)

    def addCol(self, entries, obj = 0.0, lb = 0.0, ub = None):
        """Adds a single column to the LP.

        Keyword arguments:
        entries -- a list of tuples; if p is the index of the new column, then each tuple (i, k) indicates that
        A[i][p] = k, where A is the constraint matrix and k is a nonzero entry.
        obj     -- objective coefficient (default 0.0)
        lb      -- lower bound (default 0.0)
        ub      -- upper bound (default infinity)
        """
        cdef int nnonz = len(entries)
        cdef SCIP_Real* c_coefs  = <SCIP_Real*> malloc(nnonz * sizeof(SCIP_Real))
        cdef int* c_inds = <int*>malloc(nnonz * sizeof(int))
        cdef SCIP_Real c_obj = obj
        cdef SCIP_Real c_lb = lb
        cdef SCIP_Real c_ub = ub if ub != None else self.infinity()
        cdef int c_beg = 0
        cdef int i

        for i,entry in enumerate(entries):
            c_inds[i] = entry[0]
            c_coefs[i] = entry[1]

        PY_SCIP_CALL(SCIPlpiAddCols(self.lpi, 1, &c_obj, &c_lb, &c_ub, NULL, nnonz, &c_beg, c_inds, c_coefs))

        free(c_coefs)
        free(c_inds)

    def addCols(self, entrieslist, objs = None, lbs = None, ubs = None):
        """Adds multiple columns to the LP.

        Keyword arguments:
        entrieslist -- a list of lists, where the j-th inner list contains tuples (i, k) such that A[i][p] = k,
        where A is the constraint matrix, p is the index of the j-th new column, and k is a nonzero entry.
        objs  -- objective coefficient (default 0.0)
        lbs   -- lower bounds (default 0.0)
        ubs   -- upper bounds (default infinity)
        """
        cdef int ncols = len(entrieslist)
        cdef SCIP_Real* c_objs   = <SCIP_Real*> malloc(ncols * sizeof(SCIP_Real))
        cdef SCIP_Real* c_lbs    = <SCIP_Real*> malloc(ncols * sizeof(SCIP_Real))
        cdef SCIP_Real* c_ubs    = <SCIP_Real*> malloc(ncols * sizeof(SCIP_Real))
        cdef SCIP_Real* c_coefs
        cdef int* c_inds
        cdef int* c_beg
        cdef int nnonz = sum(len(entries) for entries in entrieslist)
        cdef int i

        if nnonz > 0:
            c_coefs  = <SCIP_Real*> malloc(nnonz * sizeof(SCIP_Real))
            c_inds = <int*>malloc(nnonz * sizeof(int))
            c_beg  = <int*>malloc(ncols * sizeof(int))

            tmp = 0
            for i,entries in enumerate(entrieslist):
                c_objs[i] = objs[i] if objs != None else 0.0
                c_lbs[i] = lbs[i] if lbs != None else 0.0
                c_ubs[i] = ubs[i] if ubs != None else self.infinity()
                c_beg[i] = tmp

                for entry in entries:
                    c_inds[tmp] = entry[0]
                    c_coefs[tmp] = entry[1]
                    tmp += 1

            PY_SCIP_CALL(SCIPlpiAddCols(self.lpi, ncols, c_objs, c_lbs, c_ubs, NULL, nnonz, c_beg, c_inds, c_coefs))

            free(c_beg)
            free(c_inds)
            free(c_coefs)
        else:
            for i in range(len(entrieslist)):
                c_objs[i] = objs[i] if objs != None else 0.0
                c_lbs[i] = lbs[i] if lbs != None else 0.0
                c_ubs[i] = ubs[i] if ubs != None else self.infinity()

            PY_SCIP_CALL(SCIPlpiAddCols(self.lpi, ncols, c_objs, c_lbs, c_ubs, NULL, 0, NULL, NULL, NULL))

        free(c_ubs)
        free(c_lbs)
        free(c_objs)

    def delCols(self, firstcol, lastcol):
        """Deletes a range of columns from the LP.

        Keyword arguments:
        firstcol -- first column to delete
        lastcol  -- last column to delete
        """
        PY_SCIP_CALL(SCIPlpiDelCols(self.lpi, firstcol, lastcol))

    def addRow(self, entries, lhs=0.0, rhs=None):
        """Adds a single row to the LP.

        Keyword arguments:
        entries -- a list of tuples; if q is the index of the new row, then each tuple (j, k) indicates that
        A[q][j] = k, where A is the constraint matrix and k is a nonzero entry.
        lhs     -- left-hand side of the row (default 0.0)
        rhs     -- right-hand side of the row (default infinity)
        """
        cdef int nnonz = len(entries)
        cdef SCIP_Real* c_coefs  = <SCIP_Real*> malloc(nnonz * sizeof(SCIP_Real))
        cdef int* c_inds = <int*>malloc(nnonz * sizeof(int))
        cdef SCIP_Real c_lhs = lhs
        cdef SCIP_Real c_rhs = rhs if rhs != None else self.infinity()
        cdef int c_beg = 0
        cdef int i

        for i,entry in enumerate(entries):
            c_inds[i] = entry[0]
            c_coefs[i] = entry[1]

        PY_SCIP_CALL(SCIPlpiAddRows(self.lpi, 1, &c_lhs, &c_rhs, NULL, nnonz, &c_beg, c_inds, c_coefs))

        free(c_coefs)
        free(c_inds)

    def addRows(self, entrieslist, lhss = None, rhss = None):
        """Adds multiple rows to the LP.

        Keyword arguments:
        entrieslist -- a list of lists, where the i-th inner list contains tuples (j, k) such that A[q][j] = k,
        where A is the constraint matrix, q is the index of the i-th new row, and k is a nonzero entry.
        lhss        -- left-hand side of the row (default 0.0)
        rhss        -- right-hand side of the row (default infinity)
        """
        cdef int nrows = len(entrieslist)
        cdef SCIP_Real* c_lhss  = <SCIP_Real*> malloc(nrows * sizeof(SCIP_Real))
        cdef SCIP_Real* c_rhss  = <SCIP_Real*> malloc(nrows * sizeof(SCIP_Real))
        cdef int* c_beg  = <int*>malloc(nrows * sizeof(int))
        cdef int nnonz = sum(len(entries) for entries in entrieslist)
        cdef SCIP_Real* c_coefs = <SCIP_Real*> malloc(nnonz * sizeof(SCIP_Real))
        cdef int* c_inds = <int*>malloc(nnonz * sizeof(int))
        cdef int tmp = 0
        cdef int i

        for i,entries in enumerate(entrieslist):
            c_lhss[i] = lhss[i] if lhss != None else 0.0
            c_rhss[i] = rhss[i] if rhss != None else self.infinity()
            c_beg[i]  = tmp

            for entry in entries:
                c_inds[tmp] = entry[0]
                c_coefs[tmp] = entry[1]
                tmp += 1

        PY_SCIP_CALL(SCIPlpiAddRows(self.lpi, nrows, c_lhss, c_rhss, NULL, nnonz, c_beg, c_inds, c_coefs))

        free(c_beg)
        free(c_inds)
        free(c_coefs)
        free(c_lhss)
        free(c_rhss)

    def delRows(self, firstrow, lastrow):
        """Deletes a range of rows from the LP.

        Keyword arguments:
        firstrow -- first row to delete
        lastrow  -- last row to delete
        """
        PY_SCIP_CALL(SCIPlpiDelRows(self.lpi, firstrow, lastrow))

    def getBounds(self, firstcol = 0, lastcol = None):
        """Returns all lower and upper bounds for a range of columns.

        Keyword arguments:
        firstcol -- first column (default 0)
        lastcol  -- last column (default ncols - 1)
        """
        cdef int i

        lastcol = lastcol if lastcol != None else self.ncols() - 1

        if firstcol > lastcol:
            return None

        ncols = lastcol - firstcol + 1
        cdef SCIP_Real* c_lbs = <SCIP_Real*> malloc(ncols * sizeof(SCIP_Real))
        cdef SCIP_Real* c_ubs = <SCIP_Real*> malloc(ncols * sizeof(SCIP_Real))
        PY_SCIP_CALL(SCIPlpiGetBounds(self.lpi, firstcol, lastcol, c_lbs, c_ubs))

        lbs = []
        ubs = []

        for i in range(ncols):
            lbs.append(c_lbs[i])
            ubs.append(c_ubs[i])

        free(c_ubs)
        free(c_lbs)

        return lbs, ubs

    def getSides(self, firstrow = 0, lastrow = None):
        """Returns all left- and right-hand sides for a range of rows.

        Keyword arguments:
        firstrow -- first row (default 0)
        lastrow  -- last row (default nrows - 1)
        """
        cdef int i

        lastrow = lastrow if lastrow != None else self.nrows() - 1

        if firstrow > lastrow:
            return None

        nrows = lastrow - firstrow + 1
        cdef SCIP_Real* c_lhss = <SCIP_Real*> malloc(nrows * sizeof(SCIP_Real))
        cdef SCIP_Real* c_rhss = <SCIP_Real*> malloc(nrows * sizeof(SCIP_Real))
        PY_SCIP_CALL(SCIPlpiGetSides(self.lpi, firstrow, lastrow, c_lhss, c_rhss))

        lhss = []
        rhss = []

        for i in range(firstrow, lastrow + 1):
            lhss.append(c_lhss[i])
            rhss.append(c_rhss[i])

        free(c_rhss)
        free(c_lhss)

        return lhss, rhss

    def chgObj(self, col, obj):
        """Changes objective coefficient of a single column.

        Keyword arguments:
        col -- column to change
        obj -- new objective coefficient
        """
        cdef SCIP_Real c_obj = obj
        cdef int c_col = col

        PY_SCIP_CALL(SCIPlpiChgObj(self.lpi, 1, &c_col, &c_obj))

    def chgCoef(self, row, col, newval):
        """Changes a single coefficient in the LP.

        Keyword arguments:
        row -- row to change
        col -- column to change
        newval -- new coefficient
        """
        PY_SCIP_CALL(SCIPlpiChgCoef(self.lpi, row, col, newval))

    def chgBound(self, col, lb, ub):
        """Changes the lower and upper bound of a single column.

        Keyword arguments:
        col -- column to change
        lb  -- new lower bound
        ub  -- new upper bound
        """
        cdef SCIP_Real c_lb = lb
        cdef SCIP_Real c_ub = ub
        cdef int c_col = col

        PY_SCIP_CALL(SCIPlpiChgBounds(self.lpi, 1, &c_col, &c_lb, &c_ub))

    def chgSide(self, row, lhs, rhs):
        """Changes the left- and right-hand side of a single row.

        Keyword arguments:
        row -- row to change
        lhs -- new left-hand side
        rhs -- new right-hand side
        """
        cdef SCIP_Real c_lhs = lhs
        cdef SCIP_Real c_rhs = rhs
        cdef int c_row = row

        PY_SCIP_CALL(SCIPlpiChgSides(self.lpi, 1, &c_row, &c_lhs, &c_rhs))

    def clear(self):
        """Clears the whole LP."""
        PY_SCIP_CALL(SCIPlpiClear(self.lpi))

    def nrows(self):
        """Returns the number of rows."""
        cdef int nrows

        PY_SCIP_CALL(SCIPlpiGetNRows(self.lpi, &nrows))

        return nrows

    def ncols(self):
        """Returns the number of columns."""
        cdef int ncols

        PY_SCIP_CALL(SCIPlpiGetNCols(self.lpi, &ncols))

        return ncols

    def solve(self, dual=True):
        """Solves the current LP.

        Keyword arguments:
        dual -- use the dual or primal Simplex method (default: dual)
        """
        if dual:
            PY_SCIP_CALL(SCIPlpiSolveDual(self.lpi))
        else:
            PY_SCIP_CALL(SCIPlpiSolvePrimal(self.lpi))

        cdef SCIP_Real objval
        PY_SCIP_CALL(SCIPlpiGetObjval(self.lpi, &objval))

        return objval

    def isOptimal(self):
        """
        returns true iff LP was solved to optimality.

        Returns
        -------
        bool

        """
        return SCIPlpiIsOptimal(self.lpi)

    def getObjVal(self):
        """
        Returns the objective value of the last LP solve.
        Please note that infeasible or unbounded LPs might return unexpected results.
        """
        cdef SCIP_Real objval

        PY_SCIP_CALL(SCIPlpiGetSol(self.lpi, &objval, NULL, NULL, NULL, NULL))

        return objval

    def getPrimal(self):
        """
        Returns the primal solution of the last LP solve.
        Please note that infeasible or unbounded LPs might return unexpected results.
        """
        cdef int ncols = self.ncols()
        cdef SCIP_Real* c_primalsol = <SCIP_Real*> malloc(ncols * sizeof(SCIP_Real))
        cdef int i

        PY_SCIP_CALL(SCIPlpiGetSol(self.lpi, NULL, c_primalsol, NULL, NULL, NULL))
        primalsol = [0.0] * ncols
        for i in range(ncols):
            primalsol[i] = c_primalsol[i]
        free(c_primalsol)

        return primalsol

    def isPrimalFeasible(self):
        """Returns True iff LP is proven to be primal feasible."""
        return SCIPlpiIsPrimalFeasible(self.lpi)

    def getDual(self):
        """
        Returns the dual solution of the last LP solve.
        Please note that infeasible or unbounded LPs might return unexpected results.
        """
        cdef int nrows = self.nrows()
        cdef SCIP_Real* c_dualsol = <SCIP_Real*> malloc(nrows * sizeof(SCIP_Real))
        cdef int i

        PY_SCIP_CALL(SCIPlpiGetSol(self.lpi, NULL, NULL, c_dualsol, NULL, NULL))
        dualsol = [0.0] * nrows
        for i in range(nrows):
            dualsol[i] = c_dualsol[i]
        free(c_dualsol)

        return dualsol

    def isDualFeasible(self):
        """Returns True iff LP is proven to be dual feasible."""
        return SCIPlpiIsDualFeasible(self.lpi)

    def getPrimalRay(self):
        """Returns a primal ray if possible, None otherwise."""
        cdef int ncols
        cdef SCIP_Real* c_ray
        cdef int i

        if not SCIPlpiHasPrimalRay(self.lpi):
            return None

        ncols = self.ncols()
        c_ray = <SCIP_Real*> malloc(ncols * sizeof(SCIP_Real))

        PY_SCIP_CALL(SCIPlpiGetPrimalRay(self.lpi, c_ray))
        ray = [0.0] * ncols
        for i in range(ncols):
            ray[i] = c_ray[i]
        free(c_ray)

        return ray

    def getDualRay(self):
        """Returns a dual ray if possible, None otherwise."""
        cdef int nrows
        cdef SCIP_Real* c_ray
        cdef int i

        if not SCIPlpiHasDualRay(self.lpi):
            return None

        nrows = self.nrows()
        c_ray = <SCIP_Real*> malloc(nrows * sizeof(SCIP_Real))

        PY_SCIP_CALL(SCIPlpiGetDualfarkas(self.lpi, c_ray))
        ray = [0.0] * nrows
        for i in range(nrows):
            ray[i] = c_ray[i]
        free(c_ray)

        return ray

    def getNIterations(self):
        """Returns the number of LP iterations of the last LP solve."""
        cdef int niters

        PY_SCIP_CALL(SCIPlpiGetIterations(self.lpi, &niters))

        return niters

    def getActivity(self):
        """
        Returns the row activity vector of the last LP solve.
        Please note that infeasible or unbounded LPs might return unexpected results.
        """
        cdef int nrows = self.nrows()
        cdef SCIP_Real* c_activity = <SCIP_Real*> malloc(nrows * sizeof(SCIP_Real))
        cdef int i

        PY_SCIP_CALL(SCIPlpiGetSol(self.lpi, NULL, NULL, NULL, c_activity, NULL))

        activity = [0.0] * nrows
        for i in range(nrows):
            activity[i] = c_activity[i]

        free(c_activity)

        return activity

    def getRedcost(self):
        """
        Returns the reduced cost vector of the last LP solve.
        Please note that infeasible or unbounded LPs might return unexpected results.
        """
        cdef int ncols = self.ncols()
        cdef SCIP_Real* c_redcost = <SCIP_Real*> malloc(ncols * sizeof(SCIP_Real))
        cdef int i

        PY_SCIP_CALL(SCIPlpiGetSol(self.lpi, NULL, NULL, NULL, NULL, c_redcost))

        redcost = [0.0] * ncols
        for i in range(ncols):
            redcost[i] = c_redcost[i]

        free(c_redcost)

        return redcost

    def getBasisInds(self):
        """Returns the indices of the basic columns and rows; index i >= 0 corresponds to column i, index i < 0 to row -i-1"""
        cdef int nrows = self.nrows()
        cdef int* c_binds = <int*> malloc(nrows * sizeof(int))
        cdef int i

        PY_SCIP_CALL(SCIPlpiGetBasisInd(self.lpi, c_binds))

        binds = []
        for i in range(nrows):
            binds.append(c_binds[i])

        free(c_binds)

        return binds

    def getBase(self):
        """Returns the basis status of columns and rows.

        Status values are defined in SCIP_BASESTAT: LOWER, BASIC, UPPER, ZERO.

        Returns
        -------
        tuple of (list of int, list of int)
            Column basis statuses and row basis statuses.

        """
        cdef int ncols = self.ncols()
        cdef int nrows = self.nrows()
        cdef int* c_cstat = <int*> malloc(ncols * sizeof(int))
        cdef int* c_rstat = <int*> malloc(nrows * sizeof(int))
        cdef int i

        PY_SCIP_CALL(SCIPlpiGetBase(self.lpi, c_cstat, c_rstat))

        cstat = [c_cstat[i] for i in range(ncols)]
        rstat = [c_rstat[i] for i in range(nrows)]

        free(c_rstat)
        free(c_cstat)

        return cstat, rstat

    def setBase(self, cstat, rstat):
        """Sets the basis status of columns and rows.

        Status values are defined in SCIP_BASESTAT: LOWER, BASIC, UPPER, ZERO.

        Parameters
        ----------
        cstat : list of int
            Column basis statuses (length must equal ncols).
        rstat : list of int
            Row basis statuses (length must equal nrows).

        """
        cdef int ncols = self.ncols()
        cdef int nrows = self.nrows()
        if len(cstat) != ncols:
            raise ValueError(f"cstat has length {len(cstat)}, expected {ncols}")
        if len(rstat) != nrows:
            raise ValueError(f"rstat has length {len(rstat)}, expected {nrows}")
        cdef int* c_cstat = <int*> malloc(ncols * sizeof(int))
        cdef int* c_rstat = <int*> malloc(nrows * sizeof(int))
        cdef int i

        for i in range(ncols):
            c_cstat[i] = cstat[i]
        for i in range(nrows):
            c_rstat[i] = rstat[i]

        PY_SCIP_CALL(SCIPlpiSetBase(self.lpi, c_cstat, c_rstat))

        free(c_rstat)
        free(c_cstat)

    # Parameter Methods

    def setIntParam(self, param, value):
        """
        Set an int-valued parameter.
        If the parameter is not supported by the LP solver, KeyError will be raised.

        Parameters
        ----------
        param : SCIP_LPPARAM
            name of parameter
        value : int
            value of parameter

        """
        PY_SCIP_CALL(SCIPlpiSetIntpar(self.lpi, param, value))

    def setRealParam(self, param, value):
        """
        Set a real-valued parameter.
        If the parameter is not supported by the LP solver, KeyError will be raised.

        Parameters
        ----------
        param : SCIP_LPPARAM
            name of parameter
        value : float
            value of parameter

        """
        PY_SCIP_CALL(SCIPlpiSetRealpar(self.lpi, param, value))

    def getIntParam(self, param):
        """
        Get the value of a parameter of type int.
        If the parameter is not supported by the LP solver, KeyError will be raised.

        Parameters
        ----------
        param : SCIP_LPPARAM
            name of parameter

        Returns
        -------
        int

        """
        cdef int value

        PY_SCIP_CALL(SCIPlpiGetIntpar(self.lpi, param, &value))

        return value

    def getRealParam(self, param):
        """
        Get the value of a parameter of type float.
        If the parameter is not supported by the LP solver, KeyError will be raised.

        Parameters
        ----------
        param : SCIP_LPPARAM
            name of parameter

        Returns
        -------
        float

        """
        cdef SCIP_Real value

        PY_SCIP_CALL(SCIPlpiGetRealpar(self.lpi, param, &value))

        return value



##@file benderscut.pxi
#@brief Base class of the Benderscut Plugin
cdef class Benderscut:
    cdef public Model model
    cdef public Benders benders
    cdef public str name

    def benderscutfree(self):
        pass

    def benderscutinit(self):
        pass

    def benderscutexit(self):
        pass

    def benderscutinitsol(self):
        pass

    def benderscutexitsol(self):
        pass

    def benderscutexec(self, solution, probnumber, enfotype):
        raise NotImplementedError("benderscutexec() is a fundamental callback and should be implemented in the derived class")

cdef SCIP_RETCODE PyBenderscutCopy (SCIP* scip, SCIP_BENDERS* benders, SCIP_BENDERSCUT* benderscut) noexcept with gil:
    return SCIP_OKAY

cdef SCIP_RETCODE PyBenderscutFree (SCIP* scip, SCIP_BENDERSCUT* benderscut) noexcept with gil:
    cdef SCIP_BENDERSCUTDATA* benderscutdata
    benderscutdata = SCIPbenderscutGetData(benderscut)
    PyBenderscut = <Benderscut>benderscutdata
    PyBenderscut.benderscutfree()
    return SCIP_OKAY

cdef SCIP_RETCODE PyBenderscutInit (SCIP* scip, SCIP_BENDERSCUT* benderscut) noexcept with gil:
    cdef SCIP_BENDERSCUTDATA* benderscutdata
    benderscutdata = SCIPbenderscutGetData(benderscut)
    PyBenderscut = <Benderscut>benderscutdata
    PyBenderscut.benderscutinit()
    return SCIP_OKAY

cdef SCIP_RETCODE PyBenderscutExit (SCIP* scip, SCIP_BENDERSCUT* benderscut) noexcept with gil:
    cdef SCIP_BENDERSCUTDATA* benderscutdata
    benderscutdata = SCIPbenderscutGetData(benderscut)
    PyBenderscut = <Benderscut>benderscutdata
    PyBenderscut.benderscutexit()
    return SCIP_OKAY

cdef SCIP_RETCODE PyBenderscutInitsol (SCIP* scip, SCIP_BENDERSCUT* benderscut) noexcept with gil:
    cdef SCIP_BENDERSCUTDATA* benderscutdata
    benderscutdata = SCIPbenderscutGetData(benderscut)
    PyBenderscut = <Benderscut>benderscutdata
    PyBenderscut.benderscutinitsol()
    return SCIP_OKAY

cdef SCIP_RETCODE PyBenderscutExitsol (SCIP* scip, SCIP_BENDERSCUT* benderscut) noexcept with gil:
    cdef SCIP_BENDERSCUTDATA* benderscutdata
    benderscutdata = SCIPbenderscutGetData(benderscut)
    PyBenderscut = <Benderscut>benderscutdata
    PyBenderscut.benderscutexitsol()
    return SCIP_OKAY

cdef SCIP_RETCODE PyBenderscutExec (SCIP* scip, SCIP_BENDERS* benders, SCIP_BENDERSCUT* benderscut, SCIP_SOL* sol, int probnumber, SCIP_BENDERSENFOTYPE type, SCIP_RESULT* result) noexcept with gil:
    cdef SCIP_BENDERSCUTDATA* benderscutdata
    benderscutdata = SCIPbenderscutGetData(benderscut)
    PyBenderscut = <Benderscut>benderscutdata
    if sol == NULL:
        solution = None
    else:
        solution = Solution.create(scip, sol)
    enfotype = type
    result_dict = PyBenderscut.benderscutexec(solution, probnumber, enfotype)
    result[0] = result_dict.get("result", <SCIP_RESULT>result[0])
    return SCIP_OKAY



##@file Multidict.py
#@brief Implementation of Multidictionaries
def multidict(D):
    '''creates a multidictionary'''
    keys = list(D.keys())
    if len(keys) == 0:
        return [[]]
    try:
        N = len(D[keys[0]])
        islist = True
    except:
        N = 1
        islist = False
    dlist = [dict() for d in range(N)]
    for k in keys:
        if islist:
            for i in range(N):
                dlist[i][k] = D[k][i]
        else:
            dlist[0][k] = D[k]
    return [keys]+dlist



##@file reader.pxi
#@brief Base class of the Reader Plugin
cdef class Reader:
    cdef public Model model
    cdef public str name

    def readerfree(self):
        '''calls destructor and frees memory of reader'''
        pass

    def readerread(self, filename):
        '''calls read method of reader'''
        return {}

    def readerwrite(self, file, name, transformed, objsense, objoffset, objscale, binvars, intvars,
                    implvars, contvars, fixedvars, startnvars, conss, maxnconss, startnconss, genericnames):
        '''calls write method of reader'''
        return {}


cdef SCIP_RETCODE PyReaderCopy (SCIP* scip, SCIP_READER* reader) noexcept with gil:
    return SCIP_OKAY

cdef SCIP_RETCODE PyReaderFree (SCIP* scip, SCIP_READER* reader) noexcept with gil:
    cdef SCIP_READERDATA* readerdata
    readerdata = SCIPreaderGetData(reader)
    PyReader = <Reader>readerdata
    PyReader.readerfree()
    return SCIP_OKAY

cdef SCIP_RETCODE PyReaderRead (SCIP* scip, SCIP_READER* reader, const char* filename, SCIP_RESULT* result) noexcept with gil:
    cdef SCIP_READERDATA* readerdata
    readerdata = SCIPreaderGetData(reader)
    PyReader = <Reader>readerdata
    PyFilename = filename.decode('utf-8')
    result_dict = PyReader.readerread(PyFilename)
    result[0] = result_dict.get("result", <SCIP_RESULT>result[0])
    return SCIP_OKAY

cdef SCIP_RETCODE PyReaderWrite (SCIP* scip, SCIP_READER* reader, FILE* file,
                                 const char* filename, const char* name, SCIP_PROBDATA* probdata, SCIP_Bool transformed,
                                 SCIP_OBJSENSE objsense, SCIP_Real objoffset, SCIP_Real objscale,
                                 SCIP_RATIONAL* objoffsetexact, SCIP_RATIONAL* objscaleexact,
                                 SCIP_VAR** vars, int nvars, int nbinvars, int nintvars, int nimplvars,
                                 int ncontvars, SCIP_VAR** fixedvars, int nfixedvars, int startnvars,
                                 SCIP_CONS** conss, int nconss, int maxnconss, int startnconss,
                                 SCIP_Bool genericnames, SCIP_RESULT* result) noexcept with gil:
    cdef SCIP_READERDATA* readerdata = SCIPreaderGetData(reader)
    cdef int fd = fileno(file)
    cdef int i

    PyFile = os.fdopen(fd, "w", closefd=False)
    PyName = name.decode('utf-8')
    PyBinVars = [Variable.create(vars[i]) for i in range(nbinvars)]
    PyIntVars = [Variable.create(vars[i]) for i in range(nbinvars, nintvars)]
    PyImplVars = [Variable.create(vars[i]) for i in range(nintvars, nimplvars)]
    PyContVars = [Variable.create(vars[i]) for i in range(nimplvars, ncontvars)]
    PyFixedVars = [Variable.create(fixedvars[i]) for i in range(nfixedvars)]
    PyConss = [Constraint.create(conss[i]) for i in range(nconss)]
    PyReader = <Reader>readerdata
    #TODO: provide rational objoffsetexact and objscaleexact
    result_dict = PyReader.readerwrite(PyFile, PyName, transformed, objsense, objoffset, objscale,
                                       PyBinVars, PyIntVars, PyImplVars, PyContVars, PyFixedVars, startnvars,
                                       PyConss, maxnconss, startnconss, genericnames)
    result[0] = result_dict.get("result", <SCIP_RESULT>result[0])

    return SCIP_OKAY



##@file nodesel.pxi
#@brief Base class of the Nodesel Plugin
cdef class Nodesel:
  cdef public Model model

  def nodefree(self):
    '''frees memory of node selector'''
    pass

  def nodeinit(self):
    ''' executed after the problem is transformed. use this call to initialize node selector data.'''
    pass

  def nodeexit(self):
    '''executed before the transformed problem is freed'''
    pass

  def nodeinitsol(self):
    '''executed when the presolving is finished and the branch-and-bound process is about to begin'''
    pass

  def nodeexitsol(self):
    '''executed before the branch-and-bound process is freed'''
    pass

  def nodeselect(self):
    '''first method called in each iteration in the main solving loop. '''
    # this method needs to be implemented by the user
    return {}

  def nodecomp(self, node1, node2):
    '''
    compare two leaves of the current branching tree

    It should return the following values:

      value < 0, if node 1 comes before (is better than) node 2
      value = 0, if both nodes are equally good
      value > 0, if node 1 comes after (is worse than) node 2.
    '''
    # this method needs to be implemented by the user
    return 0


cdef SCIP_RETCODE PyNodeselCopy (SCIP* scip, SCIP_NODESEL* nodesel) noexcept with gil:
  return SCIP_OKAY

cdef SCIP_RETCODE PyNodeselFree (SCIP* scip, SCIP_NODESEL* nodesel) noexcept with gil:
  cdef SCIP_NODESELDATA* nodeseldata
  nodeseldata = SCIPnodeselGetData(nodesel)
  PyNodesel = <Nodesel>nodeseldata
  PyNodesel.nodefree()
  return SCIP_OKAY

cdef SCIP_RETCODE PyNodeselInit (SCIP* scip, SCIP_NODESEL* nodesel) noexcept with gil:
  cdef SCIP_NODESELDATA* nodeseldata
  nodeseldata = SCIPnodeselGetData(nodesel)
  PyNodesel = <Nodesel>nodeseldata
  PyNodesel.nodeinit()
  return SCIP_OKAY


cdef SCIP_RETCODE PyNodeselExit (SCIP* scip, SCIP_NODESEL* nodesel) noexcept with gil:
  cdef SCIP_NODESELDATA* nodeseldata
  nodeseldata = SCIPnodeselGetData(nodesel)
  PyNodesel = <Nodesel>nodeseldata
  PyNodesel.nodeexit()
  return SCIP_OKAY

cdef SCIP_RETCODE PyNodeselInitsol (SCIP* scip, SCIP_NODESEL* nodesel) noexcept with gil:
  cdef SCIP_NODESELDATA* nodeseldata
  nodeseldata = SCIPnodeselGetData(nodesel)
  PyNodesel = <Nodesel>nodeseldata
  PyNodesel.nodeinitsol()
  return SCIP_OKAY

cdef SCIP_RETCODE PyNodeselExitsol (SCIP* scip, SCIP_NODESEL* nodesel) noexcept with gil:
  cdef SCIP_NODESELDATA* nodeseldata
  nodeseldata = SCIPnodeselGetData(nodesel)
  PyNodesel = <Nodesel>nodeseldata
  PyNodesel.nodeexitsol()
  return SCIP_OKAY

cdef SCIP_RETCODE PyNodeselSelect (SCIP* scip, SCIP_NODESEL* nodesel, SCIP_NODE** selnode) noexcept with gil:
  cdef SCIP_NODESELDATA* nodeseldata
  nodeseldata = SCIPnodeselGetData(nodesel)
  PyNodesel = <Nodesel>nodeseldata
  result_dict = PyNodesel.nodeselect()
  selected_node = <Node>result_dict.get("selnode", None)
  if selected_node == None:
    selnode[0] = NULL
  else:
    selnode[0] = selected_node.scip_node
  
  return SCIP_OKAY

cdef int PyNodeselComp (SCIP* scip, SCIP_NODESEL* nodesel, SCIP_NODE* node1, SCIP_NODE* node2) noexcept with gil:
  cdef SCIP_NODESELDATA* nodeseldata
  nodeseldata = SCIPnodeselGetData(nodesel)
  PyNodesel = <Nodesel>nodeseldata
  n1 = Node.create(node1)
  n2 = Node.create(node2)
  result = PyNodesel.nodecomp(n1, n2) #
  return result



##@file propagator.pxi
#@brief Base class of the Propagators Plugin
cdef class Prop:
    cdef public Model model

    def propfree(self):
        '''calls destructor and frees memory of propagator'''
        pass

    def propinit(self):
        '''initializes propagator'''
        pass

    def propexit(self):
        '''calls exit method of propagator'''
        pass

    def propinitsol(self):
        '''informs propagator that the prop and bound process is being started'''
        pass

    def propexitsol(self, restart):
        '''informs propagator that the prop and bound process data is being freed'''
        pass

    def propinitpre(self):
        '''informs propagator that the presolving process is being started'''
        pass

    def propexitpre(self):
        '''informs propagator that the presolving process is finished'''
        pass

    def proppresol(self, nrounds, presoltiming, result_dict):
        '''executes presolving method of propagator'''
        pass

    def propexec(self, proptiming):
        '''calls execution method of propagator'''
        raise NotImplementedError("propexec() is a fundamental callback and should be implemented in the derived class")

    def propresprop(self, confvar, inferinfo, bdtype, relaxedbd):
        '''resolves the given conflicting bound, that was reduced by the given propagator'''
        raise NotImplementedError("propresprop() is a fundamental callback and should be implemented in the derived class")



cdef SCIP_RETCODE PyPropCopy (SCIP* scip, SCIP_PROP* prop) noexcept with gil:
    return SCIP_OKAY

cdef SCIP_RETCODE PyPropFree (SCIP* scip, SCIP_PROP* prop) noexcept with gil:
    cdef SCIP_PROPDATA* propdata
    propdata = SCIPpropGetData(prop)
    PyProp = <Prop>propdata
    PyProp.propfree()
    return SCIP_OKAY

cdef SCIP_RETCODE PyPropInit (SCIP* scip, SCIP_PROP* prop) noexcept with gil:
    cdef SCIP_PROPDATA* propdata
    propdata = SCIPpropGetData(prop)
    PyProp = <Prop>propdata
    PyProp.propinit()
    return SCIP_OKAY

cdef SCIP_RETCODE PyPropExit (SCIP* scip, SCIP_PROP* prop) noexcept with gil:
    cdef SCIP_PROPDATA* propdata
    propdata = SCIPpropGetData(prop)
    PyProp = <Prop>propdata
    PyProp.propexit()
    return SCIP_OKAY

cdef SCIP_RETCODE PyPropInitpre (SCIP* scip, SCIP_PROP* prop) noexcept with gil:
    cdef SCIP_PROPDATA* propdata
    propdata = SCIPpropGetData(prop)
    PyProp = <Prop>propdata
    PyProp.propinitpre()
    return SCIP_OKAY

cdef SCIP_RETCODE PyPropExitpre (SCIP* scip, SCIP_PROP* prop) noexcept with gil:
    cdef SCIP_PROPDATA* propdata
    propdata = SCIPpropGetData(prop)
    PyProp = <Prop>propdata
    PyProp.propexitpre()
    return SCIP_OKAY

cdef SCIP_RETCODE PyPropInitsol (SCIP* scip, SCIP_PROP* prop) noexcept with gil:
    cdef SCIP_PROPDATA* propdata
    propdata = SCIPpropGetData(prop)
    PyProp = <Prop>propdata
    PyProp.propinitsol()
    return SCIP_OKAY

cdef SCIP_RETCODE PyPropExitsol (SCIP* scip, SCIP_PROP* prop, SCIP_Bool restart) noexcept with gil:
    cdef SCIP_PROPDATA* propdata
    propdata = SCIPpropGetData(prop)
    PyProp = <Prop>propdata
    PyProp.propexitsol(restart)
    return SCIP_OKAY

cdef SCIP_RETCODE PyPropPresol (SCIP* scip, SCIP_PROP* prop, int nrounds, SCIP_PRESOLTIMING presoltiming,
                                int nnewfixedvars, int nnewaggrvars, int nnewchgvartypes, int nnewchgbds, int nnewholes,
                                int nnewdelconss, int nnewaddconss, int nnewupgdconss, int nnewchgcoefs, int nnewchgsides,
                                int* nfixedvars, int* naggrvars, int* nchgvartypes, int* nchgbds, int* naddholes,
                                int* ndelconss, int* naddconss, int* nupgdconss, int* nchgcoefs, int* nchgsides, SCIP_RESULT* result) noexcept with gil:
    cdef SCIP_PROPDATA* propdata
    propdata = SCIPpropGetData(prop)
    PyProp = <Prop>propdata
    # dictionary for input/output parameters
    result_dict = {}
    result_dict["nfixedvars"]   = nfixedvars[0]
    result_dict["naggrvars"]    = naggrvars[0]
    result_dict["nchgvartypes"] = nchgvartypes[0]
    result_dict["nchgbds"]      = nchgbds[0]
    result_dict["naddholes"]    = naddholes[0]
    result_dict["ndelconss"]    = ndelconss[0]
    result_dict["naddconss"]    = naddconss[0]
    result_dict["nupgdconss"]   = nupgdconss[0]
    result_dict["nchgcoefs"]    = nchgcoefs[0]
    result_dict["nchgsides"]    = nchgsides[0]
    result_dict["result"]       = result[0]
    PyProp.proppresol(nrounds, presoltiming,
                      nnewfixedvars, nnewaggrvars, nnewchgvartypes, nnewchgbds, nnewholes,
                      nnewdelconss, nnewaddconss, nnewupgdconss, nnewchgcoefs, nnewchgsides, result_dict)
    result[0]       = result_dict["result"]
    nfixedvars[0]   = result_dict["nfixedvars"]
    naggrvars[0]    = result_dict["naggrvars"]
    nchgvartypes[0] = result_dict["nchgvartypes"]
    nchgbds[0]      = result_dict["nchgbds"]
    naddholes[0]    = result_dict["naddholes"]
    ndelconss[0]    = result_dict["ndelconss"]
    naddconss[0]    = result_dict["naddconss"]
    nupgdconss[0]   = result_dict["nupgdconss"]
    nchgcoefs[0]    = result_dict["nchgcoefs"]
    nchgsides[0]    = result_dict["nchgsides"]
    return SCIP_OKAY

cdef SCIP_RETCODE PyPropExec (SCIP* scip, SCIP_PROP* prop, SCIP_PROPTIMING proptiming, SCIP_RESULT* result) noexcept with gil:
    cdef SCIP_PROPDATA* propdata
    propdata = SCIPpropGetData(prop)
    PyProp = <Prop>propdata
    returnvalues = PyProp.propexec(proptiming)
    result_dict = returnvalues
    result[0] = result_dict.get("result", <SCIP_RESULT>result[0])
    return SCIP_OKAY

cdef SCIP_RETCODE PyPropResProp (SCIP* scip, SCIP_PROP* prop, SCIP_VAR* infervar, int inferinfo,
                                 SCIP_BOUNDTYPE boundtype, SCIP_BDCHGIDX* bdchgidx, SCIP_Real relaxedbd, SCIP_RESULT* result) noexcept with gil:
    cdef SCIP_PROPDATA* propdata
    cdef SCIP_VAR* tmp
    tmp = infervar
    propdata = SCIPpropGetData(prop)
    confvar = Variable.create(tmp)

    #TODO: parse bdchgidx?

    PyProp = <Prop>propdata
    returnvalues = PyProp.propresprop(confvar, inferinfo, boundtype, relaxedbd)
    result_dict = returnvalues
    result[0] = result_dict.get("result", <SCIP_RESULT>result[0])
    return SCIP_OKAY



##@file heuristic.pxi
#@brief Base class of the Heuristics Plugin
cdef class Heur:
    cdef public Model model
    cdef public str name

    def heurfree(self):
        '''calls destructor and frees memory of primal heuristic'''
        pass

    def heurinit(self):
        '''initializes primal heuristic'''
        pass

    def heurexit(self):
        '''calls exit method of primal heuristic'''
        pass

    def heurinitsol(self):
        '''informs primal heuristic that the branch and bound process is being started'''
        pass

    def heurexitsol(self):
        '''informs primal heuristic that the branch and bound process data is being freed'''
        pass

    def heurexec(self, heurtiming, nodeinfeasible):
        '''should the heuristic the executed at the given depth, frequency, timing,...'''
        raise NotImplementedError("heurexec() is a fundamental callback and should be implemented in the derived class")



cdef SCIP_RETCODE PyHeurCopy (SCIP* scip, SCIP_HEUR* heur) noexcept with gil:
    return SCIP_OKAY

cdef SCIP_RETCODE PyHeurFree (SCIP* scip, SCIP_HEUR* heur) noexcept with gil:
    cdef SCIP_HEURDATA* heurdata
    heurdata = SCIPheurGetData(heur)
    PyHeur = <Heur>heurdata
    PyHeur.heurfree()
    return SCIP_OKAY

cdef SCIP_RETCODE PyHeurInit (SCIP* scip, SCIP_HEUR* heur) noexcept with gil:
    cdef SCIP_HEURDATA* heurdata
    heurdata = SCIPheurGetData(heur)
    PyHeur = <Heur>heurdata
    PyHeur.heurinit()
    return SCIP_OKAY

cdef SCIP_RETCODE PyHeurExit (SCIP* scip, SCIP_HEUR* heur) noexcept with gil:
    cdef SCIP_HEURDATA* heurdata
    heurdata = SCIPheurGetData(heur)
    PyHeur = <Heur>heurdata
    PyHeur.heurexit()
    return SCIP_OKAY

cdef SCIP_RETCODE PyHeurInitsol (SCIP* scip, SCIP_HEUR* heur) noexcept with gil:
    cdef SCIP_HEURDATA* heurdata
    heurdata = SCIPheurGetData(heur)
    PyHeur = <Heur>heurdata
    PyHeur.heurinitsol()
    return SCIP_OKAY

cdef SCIP_RETCODE PyHeurExitsol (SCIP* scip, SCIP_HEUR* heur) noexcept with gil:
    cdef SCIP_HEURDATA* heurdata
    heurdata = SCIPheurGetData(heur)
    PyHeur = <Heur>heurdata
    PyHeur.heurexitsol()
    return SCIP_OKAY

cdef SCIP_RETCODE PyHeurExec (SCIP* scip, SCIP_HEUR* heur, SCIP_HEURTIMING heurtiming, SCIP_Bool nodeinfeasible, SCIP_RESULT* result) noexcept with gil:
    cdef SCIP_HEURDATA* heurdata
    heurdata = SCIPheurGetData(heur)
    PyHeur = <Heur>heurdata
    result_dict = PyHeur.heurexec(heurtiming, nodeinfeasible)
    result[0] = result_dict.get("result", <SCIP_RESULT>result[0])
    return SCIP_OKAY
